/*
    libporty - file system related calls
    Copyright (C) 2011..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <libporty/config.h>
#include "fs.h"

#ifdef WIN32
char P_path_sep = '\\';
#else
char P_path_sep = '/';
#endif

P_filesize_t P_file_size(const char *name)
{
	struct stat buf;
	if (stat(name, &buf) != 0)
		return -1;
	return buf.st_size;
}

int P_stat(const char *path, P_stat_t *buff)
{
	struct stat buf;
	if (stat(path, &buf) != 0)
		return -1;
	buff->mode  = buf.st_mode;
	buff->nlink = buf.st_nlink;
	buff->size  = buf.st_size;
	buff->mtime = buf.st_mtime;
	return 0;
}

#ifdef WIN32
#include <windows.h>
int P_fs_owner(const char *path, int *uid, char *name, int namelen)
{
	DWORD size, len, size2;
	void *oi;
	LPTSTR name_;
	PSID SID;
	BOOL b;
	SID_NAME_USE euse;

	GetFileSecurity(path, OWNER_SECURITY_INFORMATION, NULL, 0, &size);
	oi = malloc(size);
	if (!GetFileSecurity(path, OWNER_SECURITY_INFORMATION, oi, size, &size)) {
		return -1;
	}
	if (!GetSecurityDescriptorOwner(oi, &SID, &b)) {
		return -1;
	}
	size = size2 = 0;
	name_ = NULL;
	LookupAccountSid(NULL, SID, name_, &size, NULL, &size2, &euse);

	if (uid != NULL)
		*uid = *GetSidSubAuthority(SID, *GetSidSubAuthorityCount(SID) - 1);

	if (namelen < 0)
		return 0;

	if ((name != NULL) && (namelen > 0)) {
		name_ = GlobalAlloc(GMEM_FIXED, size);
		LookupAccountSid(NULL, SID, name_, &size, NULL, &size2, &euse);
		strncpy(name, name_, namelen);
		GlobalFree(name_);
	}

	return size;
}
#else
#include <pwd.h>
int P_fs_owner(const char *path, int *uid, char *name, int namelen)
{
	struct stat buf;
	struct passwd *p;
	int len;

	if (stat(path, &buf) != 0)
		return -1;

	if (uid != NULL)
		*uid = buf.st_uid;

	if (namelen < 0)
		return 0;

	p = getpwuid(buf.st_uid);
	if (p == NULL)
		return -2;

	len = strlen(p->pw_name);
	if ((name != NULL) && (namelen > 0)) {
		strncpy(name, p->pw_name, namelen);
	}
	return len;
}
#endif

#ifdef P_HAVE_READDIR
#include "fs_includes.h"

void *P_open_dir(const char *dirname)
{
	return opendir(dirname);
}

const char *P_read_dir(void *handle)
{
	struct dirent *dp;
	
	dp = readdir((DIR *)handle);
	if (dp == NULL)
		return NULL;
	return dp->d_name;
}
 
void P_close_dir(void *handle)
{
	closedir((DIR *)handle);
}

#else
#ifdef P_HAVE_FINDNEXTFILE

typedef struct {
	HANDLE h;
	char *first;
	char *free;
} P_dirent_t;

void *P_open_dir(const char *dirname)
{
	WIN32_FIND_DATA fd;
	HANDLE h;
	P_dirent_t *p;
	char *s;

	s = malloc(strlen(dirname)+10);
	sprintf(s, "%s\\*", dirname);
	h = FindFirstFile(s, &fd);
	free(s);
	if (h == INVALID_HANDLE_VALUE)
		return NULL;
	p = malloc(sizeof(P_dirent_t));
	p->h = h;
	p->first = strdup(fd.cFileName);
	p->free = NULL;
	return p;
}

const char *P_read_dir(void *handle)
{
	WIN32_FIND_DATA fd;
	P_dirent_t *p = handle;

	if (p->first != NULL) {
		p->free  = p->first;
		p->first = NULL;
		return p->free;
	}

	if (p->free != NULL) {
		free(p->free);
		p->free = NULL;
	}

	if (FindNextFile(p->h, &fd) != 0)
		return strdup(fd.cFileName);

	return NULL;
}

void P_close_dir(void *handle)
{
	P_dirent_t *p = handle;

	if (p->first != NULL)
		free(p->first);
	else if (p->free != NULL)
		p->free = NULL;
	
	FindClose(p->h);
	free(p);
}
#else
#error no suitable readdir implementation
#endif
#endif
