/*
    libporty - IPv4 dns
    Copyright (C) 2005 gpmi-extension project
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "../config.h"
#define P_NET_HOSTNAME_SIZE 256

#ifdef WIN32
/* this define is needed for getaddrinfo, freeaddrinfo and getnameinfo
 * because mingw enclose their declarations in #if (_WIN32_WINNT >= 0x0501)
 * but they are available on all windows version (accordiing msdn)
 */
#define _WIN32_WINNT 0x0501
#endif

#include "dns4.h"

/* If return is non-zero, something went wrong! */
int P_dns4_name_to_IP(unsigned long int *ip, const char *hostname)
{
	struct addrinfo req, *ans;
	int code;

	memset(&req, 0, sizeof(req));
	req.ai_flags    = 0;
	req.ai_family   = PF_INET;
	req.ai_socktype = SOCK_STREAM;
	req.ai_protocol = IPPROTO_TCP;

	if ((code = getaddrinfo(hostname, NULL, &req, &ans)) != 0)
		return code;

	*ip = ((struct sockaddr_in *)ans->ai_addr)->sin_addr.s_addr;

	freeaddrinfo(ans);

	return 0;
}

int P_dns4_IP_to_name(char **hostname, unsigned long int *ip)
{
	struct sockaddr_in sa;
	char *buf;

	memset(&sa, 0, sizeof(sa));
	sa.sin_family      = AF_INET;
	sa.sin_addr.s_addr = *ip;

	/* we expect getnameinfo to be succesfull most of the time, so we allocate in
	   any way and then free if we failed */
	buf = malloc(P_NET_HOSTNAME_SIZE);
	if (getnameinfo((struct sockaddr *)&sa, sizeof(sa), buf, P_NET_HOSTNAME_SIZE, NULL, 0, 0)) {
		free(buf);
		return P_net_err_dns;
	}

	*hostname = buf;

	return 0;
}

char *P_dns4_get_my_name(void)
{
	char *buf;

	buf = malloc(P_NET_HOSTNAME_SIZE);
	gethostname(buf, P_NET_HOSTNAME_SIZE);

	return buf;
}

unsigned long int P_dns4_get_my_IP(void)
{
	unsigned long int ip;
	char *name = P_dns4_get_my_name();

	ip = 0xFFFFFFFF;
	P_dns4_name_to_IP(&ip, name);
	free(name);

	return ip;
}

int P_dns4_is_ip(char *name)
{
	char tmp[3];
	char *s, *d;
	int c, dots;

	*tmp = 0;

	for(s = name, c = 3, dots = 0, d = tmp;; s++) {
		if (isdigit(*s)) {
			/* None of the tags may be longer than 3 characters */
			c--;
			if (c <= 0)
				return 0;

			*d = *s;
			d++;
		}
		else if ((*s == '.') || (*s == '\0')) { /* A separator */
			c = atoi(tmp);
			/* a tag is out of range */
			if ((*tmp == '\0') || (c < 0) || (c > 255))
				return 0;

			/* Check if end of the string, if so, return true */
			if (*s == '\0')
				return 1;

			c		= 3;
			d		= tmp;
			*d	= 0;
		}
		else /* Contains a char that is neither a digit nor a dot */
			return 0;
	}

	/* We can't get here */
	return 0;
}

int P_dns4_IP_to_addr(unsigned long int *ip_in, char *addr_out)
{
	unsigned long int ip_ = htonl(*ip_in);
	unsigned char *ip = (unsigned char *)&ip_;
	char *s;
	int n;

	s = addr_out;
	for(n = 0; n < 4; n++) {
		s += sprintf(s, "%d.", ip[n]);
	}
	if (s > addr_out)
		s--;
	*s = '\0';
	return 0;
}

int P_dns4_addr_to_IP(unsigned long int *ip_out, const char *addr_in)
{
	const char *start;
	char *end;
	int token, n;
	unsigned char *ip = (unsigned char *)ip_out;

	start = addr_in;
	for(n = 0; n < 4; n++) {
		token = strtol(start, &end, 10);
		if ((*end != '.') && (*end != '\0'))
			return -1;
		if ((*end == '\0') && (n != 3))
			return -2;
		ip[3-n] = token;
		start = end + 1;
	}
	*ip_out = ntohl(*ip_out);
	return 0;
}
