/*
    libporty - stub DNS resolver
    Copyright (C) 2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include <libporty/net/network.h>
#include <libporty/net/tcp4.h>
/* for htons */
#include <arpa/inet.h>
/* for debugging */
#include <stdio.h>

#include "resolv.h"

/* Reference: RFC 1035 */

FILE *P_net_resolver_log = NULL;

static const char *TYPE_CLASS_IN = "\000\001\000\001";
static const char *TYPE_CLASS_PTR = "\000\014\000\001";

typedef enum {
	RT_A = 1,
	RT_NS = 2,
	RT_CNAME = 5,
	RT_SOA = 6,
	RT_PTR = 12,
	RT_MX = 15,
	RT_TXT = 16
} resolv_type_t;

typedef enum {
	RC_IN = 1
} resolv_class_t;

typedef struct resolv_header_s {
	unsigned short int ID;
	unsigned short int flags;
	unsigned short int qdcount, ancount, nscount, arcount;
	char data[1];
} resolv_header_t;

static char *inaddr(const char *ip)
{
	char *ret, *start, *o;
	const char *end, *last;
	int len;

#define append \
	len = last - end-1; \
	if (len <= 0) { \
		free(ret); \
		return NULL; \
	} \
	memcpy(o, end+1, len); \
	o+=len; \
	*o = '.'; \
	o++;

	len = strlen(ip);
	if (len > 15) /* 4*3 digits and 3 dots */
		return NULL;
	o = ret = malloc(len + 14);
	end = ip + len;
	for(last = end; end >= ip; end--) {
		if (*end == '.') {
			append;
			last = end;
		}
	}
	append;
	strcpy(o, "in-addr.arpa");
	return ret;
#undef append
}

static char *pack_domain(const char *hostname, char *dst)
{
	char *next;
	const char *curr;
	int len;

	if (strlen(hostname) > 255) /* RFC section 2.3.4. */
		return NULL;
	curr = hostname;
	for(;;) {
		next = strchr(curr, '.');
		if (next == NULL)
			len = strlen(curr);
		else
			len = next - curr;
		if (len > 63) /* RFC section 2.3.4. */
			return NULL;
		*dst = len;
		dst++;
		memcpy(dst, curr, len);
		dst += len;
		if (next == NULL) {
			*dst = '\0';
			dst++;
			return dst;
		}
		curr = next+1;
	}
}

static const char *unpack_domain(const char *packed, char *unpacked, const char *packet, int packet_len)
{
	int len;
	char *out;
	const char *ret, *end;
	unsigned short int ptr;

	end = packet + packet_len;

	ptr = ntohs(*((unsigned short int *)packed));
	if (ptr & 0xc000) {
		/* pointer  */
		/* make sure we return from after the 2 bytes long pointer, not from after where it points to */
		ret = packed + 2;
		/* set packed to a new location in packet */
		packed = packet+(ptr & 0x1FF);
		/* if new location is a pointer too, panic to avoid pointer loops */
		if ((*packed) & 0xc0)
			return NULL;
	}
	else
		ret = NULL;

	if ((packed >= end) || (packed < packet))
		return NULL;

	out = unpacked;
	while(*packed != 0) {
		if (packed >= end) /* don't read beyond the packet */
			return NULL;
		len = *packed;
		if (packed+len >= end) /* dont read beyond the packet */
			return NULL;
		packed++;
		memcpy(out, packed, len);
		out += len;
		*out = '.';
		out++;
		packed += len;
	}

	if (out > unpacked)
		out--;
	*out = '\0';

	/* if it was not a pointer, return where we are after reading */
	if (ret == NULL)
		ret = packed+1; /* +1 because the while loop exited right at terminating zero */
	return ret;
}

static char *unpack_ip(const char *ip_num)
{
	char *ret, *o;
	int i;

	o = ret = malloc(16);
	for(i = 0; i < 4; i++) {
		o += sprintf(o, "%d", (unsigned char)ip_num[i]);
		*o = '.';
		o++;
	}
	o--;
	*o = '\0';
	return ret;
}

char *P_resolv_asm_req(char *host_or_ip, int *length, int reverse)
{
	resolv_header_t *hdr;
	short int *req_len;
	char *req, *trail;
	int len;
	char *hostname;

	if (reverse)
		hostname = inaddr(host_or_ip);
	else
		hostname = host_or_ip;

	if (hostname == NULL) {
		if (P_net_resolver_log != NULL)
			fprintf(P_net_resolver_log, "req err: host_or_ip format\n");
		return NULL;
	}

	len = strlen(hostname) + 20;
	req = malloc(len);
	hdr = (resolv_header_t *)((char *)req + 2);
	req_len = (short int *)req;
	*req_len = htons(len-2);

	hdr->ID      = 0;
	hdr->flags   = htons(256); /* request recursion */
	hdr->qdcount = htons(1);
	hdr->ancount = 0;
	hdr->nscount = 0;
	hdr->arcount = 0;
	trail = pack_domain(hostname, hdr->data);
	if (trail == NULL) {
		if (P_net_resolver_log != NULL)
			fprintf(P_net_resolver_log, "req err: hostname format or length\n");
		goto error;
	}

	if (reverse)
		memcpy(trail, TYPE_CLASS_PTR, 4); /* type and class are hardwired for PTR */
	else
		memcpy(trail, TYPE_CLASS_IN, 4); /* type and class are hardwired for IN */
	*length = len;

	if (reverse)
		free(hostname);

	return req;

	error:;
	free(req);
	return NULL;
}

int P_resolv_decode_ans(const char *req, int len, char *result, int reverse)
{
	unsigned short int len_answ, f, dlen;
	resolv_header_t *hdr;
	const char *s, *end;
	char rdomain[256], adomain[256];

	len_answ = ntohs(*((short int *)req));
	req += 2;
	len -= 2;
	end = req + len;

	/* check message length against tcp length */
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log, "ans len: %d %d\n", len_answ, len);
	if (len_answ != len)
		return -1;

	/* message can't be shorter than the header */
	if (len < sizeof(resolv_header_t))
		return -1;

	/* decode header */
	hdr = (resolv_header_t *)req;
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans id: %d\n", ntohs(hdr->ID));
	f = ntohs(hdr->flags);
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans flags: 0x%x\n", f);
	if (!(f & 0x8000)) /* must be an answer */
		return -1;
	if (!(f & 0x0080)) /* must be a recursive answer (for now) */
		return -1;
#warning TODO: check for error flag

	if (ntohs(hdr->qdcount) != 1) /* there was exactly 1 request */
		return -1;
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans ancount: %d\n", ntohs(hdr->ancount));
	if (ntohs(hdr->ancount) < 1) /* need at least 1 answer */
		return -1;
	s = unpack_domain(hdr->data, rdomain, req, len);
	if (s >= end)
		return -1;

	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans req domain was: '%s'\n", rdomain);
	if (memcmp(s, (reverse ? TYPE_CLASS_PTR : TYPE_CLASS_IN), 4) != 0) /* type and class must be IN */
		return -1;
	s += 4; /* skip type and class */
	if (s >= end)
		return -1;

	s = unpack_domain(s, adomain, req, len);
	if (s >= end)
		return -1;
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans answer domain is: '%s'\n", adomain);
	if (memcmp(s, (reverse ? TYPE_CLASS_PTR : TYPE_CLASS_IN), 4) != 0) /* type and class must be PTR or IN */
		return -1;
	s += 4; /* skip type and class */
	s += 4; /* skip TTL */
	if (s >= end)
		return -1;

	dlen = ntohs(*((short int *)s));
	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"ans dlen=%d\n", dlen);
	if (!reverse && (dlen != 4)) /* data len must be 4 for IPv4 */
		return -1;
	s += 2;
	if (s >= end)
		return -1;

	if (reverse)
		unpack_domain(s, result, req, len);
	else {
		if (s+4 > end)
			return -1;
		memcpy(result, s, 4);
	}
	return 0;
}

/* returns 1 on authoritative negative answer, -1 for other fails */
int P_resolv_(const char *server_ip, char **host,  char **ip)
{
	P_net_socket server;
	char *req;
	char ans[512];
	int len, reverse;
	char result[256];
	int ret;

	/* get a blocking connection to the server */
	if (P_tcp4_connect(&server, server_ip, 53, NULL, 0, 0) != 0)
		return -1;

	ret = 0;

	if (*host != NULL) {
		req = P_resolv_asm_req(*host, &len, 0);
		reverse = 0;
	}
	else {
		req = P_resolv_asm_req(*ip, &len, 1);
		reverse = 1;
	}

	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"resolv: writing req\n");
	if (P_net_write(server, req, len) != len) {
		ret = -1;
		goto error;
	}

	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"resolv: reading ans\n");

	/* For now don't expect packets to be bigger than MTU */
#warning it is still brave to expect reading the whole packet in a single go - if it fails, the parser will report error
	len = P_net_read(server, ans, sizeof(ans));

	if (P_net_resolver_log != NULL)
		fprintf(P_net_resolver_log,"resolv: decoding ans\n");
	if (P_resolv_decode_ans(ans, len, result, reverse) != 0)
		return -1;

	if (!reverse) {
		*ip = unpack_ip(result);
		if (P_net_resolver_log != NULL)
			fprintf(P_net_resolver_log,"resolv: result IP is '%s'\n", *ip);
	}
	else {
		int len;
		len = strlen(result);
		*host = malloc(len+1);
		memcpy(*host, result, len+1);
		if (P_net_resolver_log != NULL)
			fprintf(P_net_resolver_log,"resolv: result host is '%s'\n", *host);
	}

	error:;
	P_net_close(server);
	free(req);
	return ret;
}

int P_resolv(char **host,  char **ip)
{
	/* bail out for syntactically wrong requests */
	if ((*host == NULL) && (*ip == NULL))
		return -1;
	if ((*host != NULL) && (*ip != NULL))
		return -1;

	return P_resolv_("8.8.8.8", host, ip);
}
