/*
    c99tree - source location handling
    Copyright (C) 2010  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "genht/htsp.h"
#include "genht/hash.h"

/* --- hash --- */
static int keyeq(char *a, char *b)
{
	return !strcmp(a, b);
}

static unsigned int hash(char *key)
{
	return murmurhash(key, strlen(key));
}

static htsp_t *loc_files = NULL;

static loc_file_t *loc_file_lookup(const char *fn, const char *path)
{
	loc_file_t *f;

	if (loc_files == NULL)
		loc_files = htsp_alloc(hash, keyeq);

	f = (loc_file_t *)htsp_get(loc_files, (htsp_key_t)path);
	if (f != NULL)
		return f;

	f = malloc(sizeof(loc_file_t));
	f->fn      = strdup(fn);
	f->path    = strdup(path);
	htsp_set(loc_files, (htsp_key_t)path, (htsp_value_t)f);
	return f;
}

void loc_files_destroy()
{
	htsp_entry_t *e;
	loc_file_t *f;

	if (loc_files == NULL)
		return;
	for (e = htsp_first(loc_files); e; e = htsp_next(loc_files, e)) {
		f = e->value;
		free(f->fn);
		free(f->path);
		free(e->value);
	}
	htsp_free(loc_files);
	loc_files = NULL;
}

static loc_t **stack = NULL;
static int alloced = 0, used = 0;
loc_t *loc_current = NULL;

void loc_file_open(const char *fn, const char *path)
{
	/* push */
	if (used == alloced) {
		alloced += 16;
		stack = realloc(stack, sizeof(loc_t) * alloced);
	}
	loc_current = stack[used] = malloc(sizeof(loc_t));
	loc_current->file = loc_file_lookup(fn, path);
	loc_current->soffs = loc_current->eoffs = 0;
	loc_current->sline = loc_current->eline = 1;
	loc_current->scol  = loc_current->ecol  = 1;
	used++;
}

void loc_file_close(void)
{
	used--;
	if (loc_current != NULL)
		free(loc_current);
	if (used > 0)
		loc_current = stack[used-1];
	else
		loc_current = NULL;
}

const char *loc_update_end(const char *token)
{
	const char *s, *ps;

	/* copy end to start */
	loc_current->soffs = loc_current->eoffs;
	loc_current->sline = loc_current->eline;
	loc_current->scol  = loc_current->ecol;

	/* update end counting lines */
	loc_current->eoffs += strlen(token);
	for(ps = s = token; (s = strchr(s, '\n')) != NULL; s++) {
		loc_current->ecol = 0;
		loc_current->eline++;
		ps = s;
	}

	/* ps is the last partial line */
	loc_current->ecol += strlen(ps);
	return token;
}

void loc_cpy(loc_t *dst, const loc_t *src)
{
	memcpy(dst, src, sizeof(loc_t));
}

static int sappend(char *at, char *end, char *what, int len)
{
	if (end - at < len)
		return -1;

	if (len == 1)
		*at = *what;
	else
		memcpy(at, what, len);

	return len;
}

static int iappend(char *at, char *end, int i)
{
	char buff[64];
	int l;

	l = sprintf(buff, "%d", i);
	return sappend(at, end, buff, l);
}

static char loc_buff[256];
char *loc_format;
const char *loc_print(const node_t *node)
{
	char *s, *f, *end;
	int len;

	if (loc_format == NULL)
		return "";

	end = loc_buff + sizeof(loc_buff) - 1;
	s   = loc_buff;
	*s  = '\0';
	for(f = loc_format; *f != '\0'; f++) {
		switch(*f) {
			case '%':
				f++;
				switch(*f) {
					case 'f': /* %f -> file name */
						len = sappend(s, end, node->loc.file->fn, strlen(node->loc.file->fn));
						break;
					case 'p': /* %p -> path name */
						len = sappend(s, end, node->loc.file->path, strlen(node->loc.file->path));
						break;
					case 'o': /* %o -> start offset */
						len = iappend(s, end, node->loc.soffs);
						break;
					case 'O': /* %O -> end offset */
						len = iappend(s, end, node->loc.eoffs);
						break;
					case 'l': /* %l -> start line */
						len = iappend(s, end, node->loc.sline);
						break;
					case 'L': /* %L -> end line */
						len = iappend(s, end, node->loc.eline);
						break;
					case 'c': /* %c -> start col */
						len = iappend(s, end, node->loc.scol);
						break;
					case 'C': /* %C -> end col */
						len = iappend(s, end, node->loc.ecol);
						break;
					case '%': /* %% -> % */
						len = sappend(s, end, f, 1);
						break;
					default: /* %* -> %* */
						len = sappend(s, end, f-1, 2);
				}
				break;
			default:
				len = sappend(s, end, f, 1);
		}
		if (len < 0)
			break;
		s += len;
	}
	*s = '\0';
	return loc_buff;
}

/* shift bytes from *in to *buff until character term found or buffer overrun */
static int sshift(const char **in, char term, char *buff, int bufflen)
{
	/* make sure we leave string empty if term is the first char */
	*buff = '\0';

	while(bufflen) {
		if (**in == term) {
			*buff = '\0';
			return 0;
		}

		/* end of string here means we didn't find term */
		if (**in == '\0') {
			fprintf(stderr, "Location shifting error: term not found\n");
			return 1;
		}

		*buff = **in;
		(*in)++;
		buff++;
		bufflen--;
	}

	fprintf(stderr, "Location shifting error: buff too short\n");
	return 1;
}

/* shift bytes from *in until character term found; convert result to integer in *i */
static int ishift(const char **in, char term, int *i)
{
	char buff[32];
	char *end;

	if (sshift(in, term, buff, sizeof(buff))) {
		fprintf(stderr, "(location shifting error while scanning an integer)\n");
		return 1;
	}

	*i = strtol(buff, &end, 0);
	if ((end == NULL) || (*end != '\0')) {
		fprintf(stderr, "Location shifting error: '%s' is not an integer\n", buff);
		return 1;
	}
	return 0;
}

/* shift a byte from *in and compare to expected */
static int cshift(const char **in, char expected)
{
	if (**in == expected) {
		(*in)++;
		return 0;
	}
	fprintf(stderr, "Location shifting error: expected %c, got %c\n", expected, **in);
	return 1;
}

static loc_t loc_scan_buff;
loc_t *loc_scan(const char *in)
{
	char *f;
	const char *i;
	char fn[512];
	char pn[512];

	*pn = *fn = '\0';
	i = in;
	for(f = loc_format; *f != '\0'; f++) {
		switch(*f) {
			case '%':
				f++;
				switch(*f) {
					case 'f': /* %f -> file name */
						if (sshift(&i, f[1], fn, sizeof(fn)))
							return NULL;
						break;
					case 'p': /* %p -> path name */
						if (sshift(&i, f[1], pn, sizeof(pn)))
							return NULL;
						break;
					case 'o': /* %o -> start offset */
						if (ishift(&i, f[1], &(loc_scan_buff.soffs)))
							return NULL;
						break;
					case 'O': /* %O -> end offset */
						if (ishift(&i, f[1], &(loc_scan_buff.eoffs)))
							return NULL;
						break;
					case 'l': /* %l -> start line */
						if (ishift(&i, f[1], &(loc_scan_buff.sline)))
							return NULL;
						break;
					case 'L': /* %L -> end line */
						if (ishift(&i, f[1], &(loc_scan_buff.eline)))
							return NULL;
						break;
					case 'c': /* %c -> start col */
						if (ishift(&i, f[1], &(loc_scan_buff.scol)))
							return NULL;
						break;
					case 'C': /* %C -> end col */
						if (ishift(&i, f[1], &(loc_scan_buff.ecol)))
							return NULL;
						break;
					case '%': /* %% -> % */
						if (cshift(&i, '%'))
							return NULL;
						break;
					default: /* %* -> %* */
						if (cshift(&i, '%'))
							return NULL;
						if (cshift(&i, *f))
							return NULL;
				}
				break;
			default:
				if (cshift(&i, *f))
					return NULL;
		}
	}
	loc_scan_buff.file = loc_file_lookup(fn, pn);
	return &loc_scan_buff;
}

void loc_destroy()
{
	loc_files_destroy();
	free(stack);
	stack = NULL;
	alloced = used = 0;
}
